<?php

namespace Drupal\pmfs;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\MessageCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Lock\LockBackendInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\pmfs\Form\Settings;

/**
 * Class Pmfs.
 *
 * @package Drupal\pmfs
 */
class Pmfs {

  use DependencySerializationTrait;
  use StringTranslationTrait;

  const KEY_TIMEOUT = 'timeout';

  const KEY_SKIP_TIMEOUT = 'skip_timeout';

  const KEY_MESSAGE = 'message';

  const KEY_STATUS = 'status';

  /**
   * Configuration factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $conf;

  /**
   * State storage service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * Locking service.
   *
   * @var \Drupal\Core\Lock\LockBackendInterface
   */
  protected $lock;

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * Static storage of loaded form settings.
   *
   * @var array[]
   */
  protected $loadedSettings;

  /**
   * Pmfs constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config
   *   Configuration factory service.
   * @param \Drupal\Core\State\StateInterface $state
   *   State storage service.
   * @param \Drupal\Core\Lock\LockBackendInterface $lock
   *   Locking service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   Current user instance.
   */
  public function __construct(
    ConfigFactoryInterface $config,
    StateInterface $state,
    LockBackendInterface $lock,
    MessengerInterface $messenger,
    AccountProxyInterface $current_user
  ) {
    $this->conf = $config;
    $this->state = $state;
    $this->lock = $lock;
    $this->messenger = $messenger;
    $this->currentUser = $current_user;
  }

  /**
   * Main controller method.
   *
   * Defines validation handler for all buttons in the form render array.
   *
   * @param array $form
   *   Form render array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   FOrm state instance.
   * @param string $form_id
   *   Form ID.
   */
  public function formAlter(array &$form, FormStateInterface $form_state, $form_id) {
    if ($form_id === Settings::FORM_ID) {
      return;
    }
    $this->devModeAlter($form, $form_state, $form_id);
    if ($this->isFormEnabled($form_id)) {
      $this->prepareCallbacks($form);
      $this->formPrepareButtons($form);
    }
  }

  /**
   * Recursively process buttons of the form.
   *
   * @param array $element
   *   Render array.
   */
  protected function formPrepareButtons(array &$element) {
    foreach ($element as $key => &$value) {
      if ($this->elementIsStructure($key, $value)) {
        $this->formPrepareButtons($value);
      }
      if ($this->elementIsButton($value)) {
        $this->prepareCallbacks($value);
      }
    }
  }

  /**
   * Check if provided element is a button.
   *
   * @param array|mixed $element
   *   Render array.
   *
   * @return bool
   *   True if element can submit a form.
   */
  protected function elementIsButton($element) {
    return is_array($element) &&
      isset($element['#type']) &&
      in_array($element['#type'], ['button', 'submit'], TRUE);
  }

  /**
   * Provides required callbacks for the submit elements.
   *
   * @param array $element
   *   Render array element to proceed.
   */
  protected function prepareCallbacks(array &$element) {
    if (!isset($element['#validate'])) {
      $element['#validate'][] = '::validateForm';
    }
    $element['#validate'] = array_merge(
      [[$this, 'formValidate']],
      $element['#validate']
    );
    if (empty($element['#submit'])) {
      // Force default submit callback.
      $element['#submit'][] = '::submitForm';
    }
    $element['#submit'][] = [$this, 'formSubmit'];
    if (isset($element['#ajax']['callback'])) {
      $element['#ajax']['pmfs_callback'] = $element['#ajax']['callback'];
      $element['#ajax']['callback'] = [$this, 'formAjaxCallback'];
    }
  }

  /**
   * Check if provided value is the structure if the render elements.
   *
   * @param string $key
   *   Key value of render array.
   * @param mixed $value
   *   Value of array item.
   *
   * @return bool
   *   Flag is value is an array.
   */
  protected function elementIsStructure($key, $value) {
    return (strpos($key, '#') !== 0) &&
      is_array($value);
  }

  /**
   * Validate form handler.
   *
   * @param array $form
   *   Form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   */
  public function formValidate(array $form, FormStateInterface $form_state) {
    // Check lock.
    if (!$this->lockMayBeAvailableByFormState($form_state)) {
      $this->setLockValidationError($form, $form_state);
      return;
    }
    // Set lock.
    $this->setLockByFormState($form_state);
  }

  /**
   * Validate form handler.
   *
   * @param string $id
   *   Request id.
   * @param bool $show_message_on_deny
   *   Flag to show error message or not.
   *
   * @return bool
   *   Flag of set lock or not.
   */
  public function isCustomRequestExecutable($id, $show_message_on_deny = TRUE) {
    // Check lock.
    if (!$this->lockMayBeAvailableByCustomId($id)) {
      if ($show_message_on_deny) {
        $this->setCustomLockErrorMessage($id);
      }
      return FALSE;
    }
    // Set lock.
    $this->setLockByCustomId($id);
    return TRUE;
  }

  /**
   * Form submit hanfler.
   *
   * @param array $form
   *   Form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   */
  public function formSubmit(array $form, FormStateInterface $form_state) {
    // Check lock.
    if ($this->forceReleaseFormLock($form_state) && !$this->lockMayBeAvailableByFormState($form_state)) {
      // Release lock.
      $this->releaseFormLock($form_state);
    }
  }

  /**
   * Custom request finished handler.
   *
   * @param string $id
   *   Custom request ID.
   */
  public function setCustomRequestDone($id) {
    // Check lock.
    if ($this->forceReleaseCustomLock($id) && !$this->lockMayBeAvailableByCustomId($id)) {
      // Release lock.
      $this->releaseCustomLock($id);
    }
  }

  /**
   * Ajax callback handler.
   *
   * @param array $form
   *   Form aarray.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   *
   * @return mixed
   *   Ajax callback result.
   */
  public function formAjaxCallback(array $form, FormStateInterface $form_state) {
    $this->formSubmit($form, $form_state);
    // Check lock.
    if (!$this->lockMayBeAvailableByFormState($form_state)) {
      $this->setLockValidationError($form, $form_state);
      $commands = new AjaxResponse();
      $commands->addCommand(new MessageCommand($this->getFormSettings($form_state)[static::KEY_MESSAGE], NULL, ['type' => 'error']));
      return $commands;
    }
    // Set lock.
    $this->setLockByFormState($form_state);
    $callback = $form_state->getTriggeringElement()['#ajax']['pmfs_callback'];
    if (strpos($callback, '::') === 0) {
      $callback = [
        $form_state->getFormObject(),
        str_replace('::', '', $callback),
      ];
    }
    return $callback($form, $form_state);
  }

  /**
   * Is form configured for force release option.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   *
   * @return bool
   *   Flag of force release option of specific form configuration.
   */
  public function forceReleaseFormLock(FormStateInterface $form_state) {
    return (bool) $this->getFormSettings($form_state)[static::KEY_SKIP_TIMEOUT];
  }

  /**
   * Is custom settings configured for force release option.
   *
   * @param string $id
   *   Form state.
   *
   * @return bool
   *   Flag of force release option of specific custom configuration.
   */
  public function forceReleaseCustomLock($id) {
    return (bool) $this->getCustomSettings($id)[static::KEY_SKIP_TIMEOUT];
  }

  /**
   * Releases the given lock for the specific form.
   *
   * @param \Drupal\Core\Form\FormStateInterface|string $id
   *   Form state or form ID.
   *
   * @see \Drupal\Core\Lock\LockBackendInterface::release()
   */
  public function releaseFormLock($id) {
    $this->lock->release(
      $this->getLockKeyByFormState($id)
    );
  }

  /**
   * Releases the given lock for the specific custom request.
   *
   * @param string $id
   *   Form state or form ID.
   *
   * @see \Drupal\Core\Lock\LockBackendInterface::release()
   */
  public function releaseCustomLock($id) {
    $this->lock->release(
      $this->getCustomLockKey($id)
    );
  }

  /**
   * Validation error message handler for the specific form configuration.
   *
   * @param array $form
   *   Form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   */
  public function setLockValidationError(array $form, FormStateInterface $form_state) {
    $form_state->setError(
      $form,
      $this->getFormSettings($form_state)[static::KEY_MESSAGE]
    );
  }

  /**
   * Validation error message handler for the specific custom configuration.
   *
   * @param string $id
   *   Lock id.
   */
  public function setCustomLockErrorMessage($id) {
    $this->messenger->addError(
      $this->getCustomSettings($id)[static::KEY_MESSAGE]
    );
  }

  /**
   * Acquires a lock for the specific form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   *
   * @see \Drupal\Core\Lock\LockBackendInterface::acquire()
   */
  public function setLockByFormState(FormStateInterface $form_state) {
    $this->lock->acquire(
      $this->getLockKeyByFormState($form_state),
      $this->getFormSettings($form_state)[static::KEY_TIMEOUT]
    );
  }

  /**
   * Acquires a lock for the specific custom request.
   *
   * @param string $id
   *   Form state.
   *
   * @see \Drupal\Core\Lock\LockBackendInterface::acquire()
   */
  public function setLockByCustomId($id) {
    $this->lock->acquire(
      $this->getCustomLockKey($id),
      $this->getCustomSettings($id)[static::KEY_TIMEOUT]
    );
  }

  /**
   * Default values getter.
   *
   * @return array
   *   Default values.
   */
  public function getDefaultValues() {
    $default_config = $this->conf->get('pmfs.settings')->get('form._global') ?? [];
    return [
      static::KEY_STATUS => (bool) ($default_config[static::KEY_STATUS] ?? FALSE),
      static::KEY_TIMEOUT => (int) ($default_config[static::KEY_TIMEOUT] ?? 30),
      static::KEY_SKIP_TIMEOUT => (bool) ($default_config[static::KEY_SKIP_TIMEOUT] ?? FALSE),
      static::KEY_MESSAGE => $default_config[static::KEY_MESSAGE] ?? $this->t('The form still processing, please, try again later.'),
    ];
  }

  /**
   * Checks if a lock is available for acquiring for the specific form.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   *
   * @return bool
   *   Lock availability flag.
   * @see \Drupal\Core\Lock\LockBackendInterface::lockMayBeAvailable()
   */
  protected function lockMayBeAvailableByFormState(FormStateInterface $form_state) {
    return $this->lock->lockMayBeAvailable(
      $this->getLockKeyByFormState($form_state)
    );
  }

  /**
   * Checks if a lock is available for acquiring for the specific form.
   *
   * @param string $id
   *   Form state.
   *
   * @return bool
   *   Lock availability flag.
   * @see \Drupal\Core\Lock\LockBackendInterface::lockMayBeAvailable()
   */
  protected function lockMayBeAvailableByCustomId($id) {
    return $this->lock->lockMayBeAvailable(
      $this->getCustomLockKey($id)
    );
  }

  /**
   * Get ID for the current user.
   *
   * @param string $type
   *   Type of the lock.
   * @param string $id
   *   ID of the lock.
   *
   * @return string
   *   Lock key.
   */
  public function getLockKey($type, $id) {
    if ($this->currentUser->isAnonymous()) {
      $user_id_sid = 'u' . session_id();
    }
    else {
      $user_id_sid = 'a' . $this->currentUser->id();
    }
    return implode('.', ['pmfs', $type, $user_id_sid, $id]);
  }

  /**
   * Get ID for the current user.
   *
   * @param string|\Drupal\Core\Form\FormStateInterface $id
   *   Form state.
   *
   * @return string
   *   Lock key.
   */
  public function getLockKeyByFormState($id) {
    if ($id instanceof FormStateInterface) {
      return $this->getLockKeyByFormState($id->getFormObject()->getFormId());
    }
    return $this->getLockKey('form', $id);
  }

  /**
   * Get custom ID for the current user.
   *
   * @param string $id
   *   Form state.
   *
   * @return string
   *   Lock key.
   */
  public function getCustomLockKey($id) {
    return $this->getLockKey('custom', $id);
  }

  /**
   * Form settings getter.
   *
   * @param string|\Drupal\Core\Form\FormStateInterface $form_id
   *   Form ID or Form State instance.
   *
   * @return array
   *   Default configs.
   */
  public function getFormSettings($form_id) {
    if ($form_id instanceof FormStateInterface) {
      return $this->getFormSettings($form_id->getFormObject()->getFormId());
    }
    if (!isset($this->loadedSettings['form'][$form_id])) {
      $settings = $this->conf
        ->get('pmfs.settings')
        ->get('form.' . $form_id) ?? [];
      $settings = array_replace($this->getDefaultValues(), $settings);
      $this->loadedSettings['form'][$form_id] = $settings;
    }
    return $this->loadedSettings['form'][$form_id];
  }

  /**
   * Custom settings getter.
   *
   * @param string $id
   *   Form ID or Form State instance.
   *
   * @return array
   *   Default configs.
   */
  public function getCustomSettings($id) {
    if (!isset($this->loadedSettings['custom'][$id])) {
      $settings = $this->conf
        ->get('pmfs.settings')
        ->get('custom.' . $id) ?? [];
      $create = empty($settings);
      $settings = array_replace($this->getDefaultValues(), $settings);
      if ($create && $this->state->get('pmfs.dev_mode.enabled', FALSE)) {
        $this->conf
          ->getEditable('pmfs.settings')
          ->set('custom.' . $id, $settings);
      }
      $this->loadedSettings['custom'][$id] = $settings;
    }
    return $this->loadedSettings['custom'][$id];
  }

  /**
   * Get flag of form configuration status.
   *
   * @param string $form_id
   *   Form ID.
   *
   * @return bool
   *   Status flag.
   */
  public function isFormEnabled($form_id) {
    return (bool) $this->getFormSettings($form_id)[static::KEY_STATUS];
  }

  /**
   * Get flag of custom configuration status.
   *
   * @param string $id
   *   Custom request ID.
   *
   * @return bool
   *   Status flag.
   */
  public function isCustomEnabled($id) {
    return (bool) $this->getCustomSettings($id)[static::KEY_STATUS];
  }

  /**
   * Dev mode features.
   *
   * @param array $form
   *   Form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   Form state.
   * @param string $form_id
   *   Form ID.
   */
  protected function devModeAlter(array $form, FormStateInterface $form_state, string $form_id) {
    if ($this->state->get('pmfs.dev_mode.enabled', FALSE)) {
      $list = $this->state->get('pmfs.dev_mode.detected_form_ids', []);
      $list[] = $form_id;
      $this->state->set('pmfs.dev_mode.detected_form_ids', array_unique($list));
    }
  }

}
